import concurrent.futures
import json
from multiprocessing import cpu_count
import os
from pathlib import Path
from setuptools import setup, find_packages
import subprocess
import sys
from typing import Any, Dict, List, Optional, Tuple

VERSION_MODULE_PATH = os.path.join(os.path.realpath(os.path.dirname(__file__)), "polyfile", "version.py")


# Build the entire Kaitai struct format library:
POLYFILE_DIR: Path = Path(__file__).absolute().parent
COMPILE_SCRIPT: Path = POLYFILE_DIR / "polyfile" / "kaitai" / "compiler.py"
KAITAI_FORMAT_LIBRARY: Path = POLYFILE_DIR / "kaitai_struct_formats"
KAITAI_PARSERS_DIR: Path = POLYFILE_DIR / "polyfile" / "kaitai" / "parsers"
MANIFEST_PATH: Path = KAITAI_PARSERS_DIR / "manifest.json"
README_PATH: Path = POLYFILE_DIR / "README.md"


# Make sure the ktaitai_struct_formats submodlue is cloned:
if not (KAITAI_FORMAT_LIBRARY / "README.md").exists():
    subprocess.check_call(["git", "submodule", "init"], cwd=str(POLYFILE_DIR))
    subprocess.check_call(["git", "submodule", "update"], cwd=str(POLYFILE_DIR))


def compile_ksy(path: Path) -> List[Tuple[str, str]]:
    output = subprocess.check_output(
        [sys.executable, str(COMPILE_SCRIPT), str(path), str(KAITAI_PARSERS_DIR)],
        cwd=str(KAITAI_FORMAT_LIBRARY)
    ).decode("utf-8")
    return [  # type: ignore
        tuple(line.split("\t")[:2])  # (class_name, python_path)
        for line in output.split("\n") if line.strip()
    ]


# see if any of the files are out of date and need to be recompiled
newest_definition: Optional[float] = None
for definition in KAITAI_FORMAT_LIBRARY.glob("**/*.ksy"):
    mtime = definition.stat().st_mtime
    if newest_definition is None or newest_definition < mtime:
        newest_definition = mtime

if not MANIFEST_PATH.exists() or newest_definition > MANIFEST_PATH.stat().st_mtime:
    # the definitions have been updated, so we need to recompile everything

    num_files = sum(1 for _ in KAITAI_FORMAT_LIBRARY.glob("**/*.ksy"))

    try:
        from tqdm import tqdm
    except ModuleNotFoundError:
        def tqdm(*args, **kwargs):
            class TQDM:
                def __enter__(self):
                    return self
                def __exit__(self, exc_type, exc_val, exc_tb):
                    pass
                def write(self, message, *args, **kwargs):
                    sys.stderr.write(message)
                    sys.stderr.write("\n")
                    sys.stderr.flush()
                def update(self, n: int):
                    pass
            return TQDM()


    ksy_manifest: Dict[str, Dict[str, Any]] = {}

    with tqdm(leave=False, desc="Compiling the Kaitai Struct Format Library", total=num_files) as t:
        with concurrent.futures.ThreadPoolExecutor(max_workers=cpu_count()) as executor:
            futures_to_path: Dict[concurrent.futures.Future, Path] = {
                executor.submit(compile_ksy, file): file
                for file in KAITAI_FORMAT_LIBRARY.glob("**/*.ksy")
            }
            for future in concurrent.futures.as_completed(futures_to_path):
                t.update(1)
                path = futures_to_path[future]
                relative_path = str(path.relative_to(KAITAI_FORMAT_LIBRARY))
                if relative_path in ksy_manifest:
                    raise ValueError(f"{relative_path} appears twice in the Kaitai format library!")
                try:
                    (first_spec_class_name, first_spec_python_path), *dependencies = future.result()
                    ksy_manifest[relative_path] = {
                        "class_name": first_spec_class_name,
                        "python_path": str(Path(first_spec_python_path).relative_to(KAITAI_PARSERS_DIR)),
                        "dependencies": [
                            {
                                "class_name": class_name,
                                "python_path": str(Path(python_path).relative_to(KAITAI_PARSERS_DIR))
                            }
                            for class_name, python_path in dependencies
                        ]
                    }
                    t.write(f"Compiled {path.name}")
                except Exception as e:
                    t.write(f"Warning: Failed to compile {path}: {e}\n")

    with open(MANIFEST_PATH, "w") as f:
        json.dump(ksy_manifest, f)


def get_version_string():
    version = {}
    with open(VERSION_MODULE_PATH) as f:
        exec(f.read(), version)
    return version['VERSION_STRING']


with open(README_PATH, "r") as readme:
    README = readme.read()

setup(
    name='polyfile',
    description='A utility to recursively map the structure of a file.',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/trailofbits/polyfile',
    author='Trail of Bits',
    version=get_version_string(),
    packages=find_packages(exclude=("tests",)),
    python_requires='>=3.6',
    install_requires=[
        "dataclasses;python_version<'3.7'",  # dataclasses were only added in Python 3.7
        'graphviz',
        'intervaltree',
        'jinja2',
        'kaitaistruct>=0.7',
        'networkx',
        'Pillow>=5.0.0',
        'pyyaml>=3.13',
        'setuptools'
    ],
    extras_require={
        'demangle': ['cxxfilt'],
        "dev": ["mypy", "pytest", "flake8"]
    },
    entry_points={
        'console_scripts': [
            'polyfile = polyfile.__main__:main',
            'polymerge = polymerge.__main__:main'
        ]
    },
    package_data={"polyfile": [
        os.path.join("templates", "*"),
        os.path.join("kaitai", "parsers", "*.py"),
        os.path.join("kaitai", "parsers", "manifest.json"),
        os.path.join("magic_defs", "*")
    ]},
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Security',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities'
    ]
)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Wmf(KaitaiStruct):
    """WMF (Windows Metafile) is a relatively early vector image format
    introduced for Microsoft Windows in 1990.
    
    Inside, it provides a serialized list of Windows GDI (Graphics
    Device Interface) function calls, which, if played back, result in
    an image being drawn on a given surface (display, off-screen buffer,
    printer, etc).
    
    .. seealso::
       Source - http://www.digitalpreservation.gov/formats/digformatspecs/WindowsMetafileFormat(wmf)Specification.pdf
    """

    class Func(Enum):
        eof = 0
        savedc = 30
        realizepalette = 53
        setpalentries = 55
        createpalette = 247
        setbkmode = 258
        setmapmode = 259
        setrop2 = 260
        setrelabs = 261
        setpolyfillmode = 262
        setstretchbltmode = 263
        settextcharextra = 264
        restoredc = 295
        invertregion = 298
        paintregion = 299
        selectclipregion = 300
        selectobject = 301
        settextalign = 302
        resizepalette = 313
        dibcreatepatternbrush = 322
        setlayout = 329
        deleteobject = 496
        createpatternbrush = 505
        setbkcolor = 513
        settextcolor = 521
        settextjustification = 522
        setwindoworg = 523
        setwindowext = 524
        setviewportorg = 525
        setviewportext = 526
        offsetwindoworg = 527
        offsetviewportorg = 529
        lineto = 531
        moveto = 532
        offsetcliprgn = 544
        fillregion = 552
        setmapperflags = 561
        selectpalette = 564
        createpenindirect = 762
        createfontindirect = 763
        createbrushindirect = 764
        polygon = 804
        polyline = 805
        scalewindowext = 1040
        scaleviewportext = 1042
        excludecliprect = 1045
        intersectcliprect = 1046
        ellipse = 1048
        floodfill = 1049
        rectangle = 1051
        setpixel = 1055
        frameregion = 1065
        animatepalette = 1078
        textout = 1313
        polypolygon = 1336
        extfloodfill = 1352
        roundrect = 1564
        patblt = 1565
        escape = 1574
        createregion = 1791
        arc = 2071
        pie = 2074
        chord = 2096
        bitblt = 2338
        dibbitblt = 2368
        exttextout = 2610
        stretchblt = 2851
        dibstretchblt = 2881
        setdibtodev = 3379
        stretchdib = 3907

    class BinRasterOp(Enum):
        black = 1
        notmergepen = 2
        masknotpen = 3
        notcopypen = 4
        maskpennot = 5
        not = 6
        xorpen = 7
        notmaskpen = 8
        maskpen = 9
        notxorpen = 10
        nop = 11
        mergenotpen = 12
        copypen = 13
        mergepennot = 14
        mergepen = 15
        white = 16

    class MixMode(Enum):
        transparent = 1
        opaque = 2

    class PolyFillMode(Enum):
        alternate = 1
        winding = 2
    SEQ_FIELDS = ["special_header", "header", "records"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['special_header']['start'] = self._io.pos()
        self.special_header = Wmf.SpecialHeader(self._io, self, self._root)
        self.special_header._read()
        self._debug['special_header']['end'] = self._io.pos()
        self._debug['header']['start'] = self._io.pos()
        self.header = Wmf.Header(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['records']['start'] = self._io.pos()
        self.records = []
        i = 0
        while True:
            if not 'arr' in self._debug['records']:
                self._debug['records']['arr'] = []
            self._debug['records']['arr'].append({'start': self._io.pos()})
            _t_records = Wmf.Record(self._io, self, self._root)
            _t_records._read()
            _ = _t_records
            self.records.append(_)
            self._debug['records']['arr'][len(self.records) - 1]['end'] = self._io.pos()
            if _.function == Wmf.Func.eof:
                break
            i += 1
        self._debug['records']['end'] = self._io.pos()

    class ParamsSetwindoworg(KaitaiStruct):
        """
        .. seealso::
           section 2.3.5.31
        """
        SEQ_FIELDS = ["y", "x"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['y']['start'] = self._io.pos()
            self.y = self._io.read_s2le()
            self._debug['y']['end'] = self._io.pos()
            self._debug['x']['start'] = self._io.pos()
            self.x = self._io.read_s2le()
            self._debug['x']['end'] = self._io.pos()


    class ParamsSetbkmode(KaitaiStruct):
        """
        .. seealso::
           section 2.3.5.15
        """
        SEQ_FIELDS = ["bk_mode"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['bk_mode']['start'] = self._io.pos()
            self.bk_mode = KaitaiStream.resolve_enum(Wmf.MixMode, self._io.read_u2le())
            self._debug['bk_mode']['end'] = self._io.pos()


    class PointS(KaitaiStruct):
        """
        .. seealso::
           section 2.2.1.12
        """
        SEQ_FIELDS = ["x", "y"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['x']['start'] = self._io.pos()
            self.x = self._io.read_s2le()
            self._debug['x']['end'] = self._io.pos()
            self._debug['y']['start'] = self._io.pos()
            self.y = self._io.read_s2le()
            self._debug['y']['end'] = self._io.pos()


    class ParamsSetwindowext(KaitaiStruct):
        """
        .. seealso::
           section 2.3.5.30
        """
        SEQ_FIELDS = ["y", "x"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['y']['start'] = self._io.pos()
            self.y = self._io.read_s2le()
            self._debug['y']['end'] = self._io.pos()
            self._debug['x']['start'] = self._io.pos()
            self.x = self._io.read_s2le()
            self._debug['x']['end'] = self._io.pos()


    class ParamsPolygon(KaitaiStruct):
        """
        .. seealso::
           section 2.3.3.15 = params_polyline
        """
        SEQ_FIELDS = ["num_points", "points"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_points']['start'] = self._io.pos()
            self.num_points = self._io.read_s2le()
            self._debug['num_points']['end'] = self._io.pos()
            self._debug['points']['start'] = self._io.pos()
            self.points = [None] * (self.num_points)
            for i in range(self.num_points):
                if not 'arr' in self._debug['points']:
                    self._debug['points']['arr'] = []
                self._debug['points']['arr'].append({'start': self._io.pos()})
                _t_points = Wmf.PointS(self._io, self, self._root)
                _t_points._read()
                self.points[i] = _t_points
                self._debug['points']['arr'][i]['end'] = self._io.pos()

            self._debug['points']['end'] = self._io.pos()


    class Header(KaitaiStruct):

        class MetafileType(Enum):
            memory_metafile = 1
            disk_metafile = 2
        SEQ_FIELDS = ["metafile_type", "header_size", "version", "size", "number_of_objects", "max_record", "number_of_members"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['metafile_type']['start'] = self._io.pos()
            self.metafile_type = KaitaiStream.resolve_enum(Wmf.Header.MetafileType, self._io.read_u2le())
            self._debug['metafile_type']['end'] = self._io.pos()
            self._debug['header_size']['start'] = self._io.pos()
            self.header_size = self._io.read_u2le()
            self._debug['header_size']['end'] = self._io.pos()
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u2le()
            self._debug['version']['end'] = self._io.pos()
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['number_of_objects']['start'] = self._io.pos()
            self.number_of_objects = self._io.read_u2le()
            self._debug['number_of_objects']['end'] = self._io.pos()
            self._debug['max_record']['start'] = self._io.pos()
            self.max_record = self._io.read_u4le()
            self._debug['max_record']['end'] = self._io.pos()
            self._debug['number_of_members']['start'] = self._io.pos()
            self.number_of_members = self._io.read_u2le()
            self._debug['number_of_members']['end'] = self._io.pos()


    class ColorRef(KaitaiStruct):
        """
        .. seealso::
           section 2.2.1.7
        """
        SEQ_FIELDS = ["red", "green", "blue", "reserved"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['red']['start'] = self._io.pos()
            self.red = self._io.read_u1()
            self._debug['red']['end'] = self._io.pos()
            self._debug['green']['start'] = self._io.pos()
            self.green = self._io.read_u1()
            self._debug['green']['end'] = self._io.pos()
            self._debug['blue']['start'] = self._io.pos()
            self.blue = self._io.read_u1()
            self._debug['blue']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_u1()
            self._debug['reserved']['end'] = self._io.pos()


    class ParamsSetrop2(KaitaiStruct):
        """
        .. seealso::
           section 2.3.5.22
        """
        SEQ_FIELDS = ["draw_mode"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['draw_mode']['start'] = self._io.pos()
            self.draw_mode = KaitaiStream.resolve_enum(Wmf.BinRasterOp, self._io.read_u2le())
            self._debug['draw_mode']['end'] = self._io.pos()


    class ParamsSetpolyfillmode(KaitaiStruct):
        """
        .. seealso::
           section 2.3.5.20
        """
        SEQ_FIELDS = ["poly_fill_mode"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['poly_fill_mode']['start'] = self._io.pos()
            self.poly_fill_mode = KaitaiStream.resolve_enum(Wmf.PolyFillMode, self._io.read_u2le())
            self._debug['poly_fill_mode']['end'] = self._io.pos()


    class ParamsPolyline(KaitaiStruct):
        """
        .. seealso::
           section 2.3.3.14
        """
        SEQ_FIELDS = ["num_points", "points"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_points']['start'] = self._io.pos()
            self.num_points = self._io.read_s2le()
            self._debug['num_points']['end'] = self._io.pos()
            self._debug['points']['start'] = self._io.pos()
            self.points = [None] * (self.num_points)
            for i in range(self.num_points):
                if not 'arr' in self._debug['points']:
                    self._debug['points']['arr'] = []
                self._debug['points']['arr'].append({'start': self._io.pos()})
                _t_points = Wmf.PointS(self._io, self, self._root)
                _t_points._read()
                self.points[i] = _t_points
                self._debug['points']['arr'][i]['end'] = self._io.pos()

            self._debug['points']['end'] = self._io.pos()


    class SpecialHeader(KaitaiStruct):
        SEQ_FIELDS = ["magic", "handle", "left", "top", "right", "bottom", "inch", "reserved", "checksum"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\xD7\xCD\xC6\x9A":
                raise kaitaistruct.ValidationNotEqualError(b"\xD7\xCD\xC6\x9A", self.magic, self._io, u"/types/special_header/seq/0")
            self._debug['handle']['start'] = self._io.pos()
            self.handle = self._io.read_bytes(2)
            self._debug['handle']['end'] = self._io.pos()
            if not self.handle == b"\x00\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x00\x00", self.handle, self._io, u"/types/special_header/seq/1")
            self._debug['left']['start'] = self._io.pos()
            self.left = self._io.read_s2le()
            self._debug['left']['end'] = self._io.pos()
            self._debug['top']['start'] = self._io.pos()
            self.top = self._io.read_s2le()
            self._debug['top']['end'] = self._io.pos()
            self._debug['right']['start'] = self._io.pos()
            self.right = self._io.read_s2le()
            self._debug['right']['end'] = self._io.pos()
            self._debug['bottom']['start'] = self._io.pos()
            self.bottom = self._io.read_s2le()
            self._debug['bottom']['end'] = self._io.pos()
            self._debug['inch']['start'] = self._io.pos()
            self.inch = self._io.read_u2le()
            self._debug['inch']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bytes(4)
            self._debug['reserved']['end'] = self._io.pos()
            if not self.reserved == b"\x00\x00\x00\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x00\x00\x00\x00", self.reserved, self._io, u"/types/special_header/seq/7")
            self._debug['checksum']['start'] = self._io.pos()
            self.checksum = self._io.read_u2le()
            self._debug['checksum']['end'] = self._io.pos()


    class Record(KaitaiStruct):
        SEQ_FIELDS = ["size", "function", "params"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['function']['start'] = self._io.pos()
            self.function = KaitaiStream.resolve_enum(Wmf.Func, self._io.read_u2le())
            self._debug['function']['end'] = self._io.pos()
            self._debug['params']['start'] = self._io.pos()
            _on = self.function
            if _on == Wmf.Func.setbkmode:
                self._raw_params = self._io.read_bytes(((self.size - 3) * 2))
                _io__raw_params = KaitaiStream(BytesIO(self._raw_params))
                self.params = Wmf.ParamsSetbkmode(_io__raw_params, self, self._root)
                self.params._read()
            elif _on == Wmf.Func.polygon:
                self._raw_params = self._io.read_bytes(((self.size - 3) * 2))
                _io__raw_params = KaitaiStream(BytesIO(self._raw_params))
                self.params = Wmf.ParamsPolygon(_io__raw_params, self, self._root)
                self.params._read()
            elif _on == Wmf.Func.setbkcolor:
                self._raw_params = self._io.read_bytes(((self.size - 3) * 2))
                _io__raw_params = KaitaiStream(BytesIO(self._raw_params))
                self.params = Wmf.ColorRef(_io__raw_params, self, self._root)
                self.params._read()
            elif _on == Wmf.Func.setpolyfillmode:
                self._raw_params = self._io.read_bytes(((self.size - 3) * 2))
                _io__raw_params = KaitaiStream(BytesIO(self._raw_params))
                self.params = Wmf.ParamsSetpolyfillmode(_io__raw_params, self, self._root)
                self.params._read()
            elif _on == Wmf.Func.setwindoworg:
                self._raw_params = self._io.read_bytes(((self.size - 3) * 2))
                _io__raw_params = KaitaiStream(BytesIO(self._raw_params))
                self.params = Wmf.ParamsSetwindoworg(_io__raw_params, self, self._root)
                self.params._read()
            elif _on == Wmf.Func.setrop2:
                self._raw_params = self._io.read_bytes(((self.size - 3) * 2))
                _io__raw_params = KaitaiStream(BytesIO(self._raw_params))
                self.params = Wmf.ParamsSetrop2(_io__raw_params, self, self._root)
                self.params._read()
            elif _on == Wmf.Func.setwindowext:
                self._raw_params = self._io.read_bytes(((self.size - 3) * 2))
                _io__raw_params = KaitaiStream(BytesIO(self._raw_params))
                self.params = Wmf.ParamsSetwindowext(_io__raw_params, self, self._root)
                self.params._read()
            elif _on == Wmf.Func.polyline:
                self._raw_params = self._io.read_bytes(((self.size - 3) * 2))
                _io__raw_params = KaitaiStream(BytesIO(self._raw_params))
                self.params = Wmf.ParamsPolyline(_io__raw_params, self, self._root)
                self.params._read()
            else:
                self.params = self._io.read_bytes(((self.size - 3) * 2))
            self._debug['params']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class OpenpgpMessage(KaitaiStruct):
    """The OpenPGP Message Format is a format to store encryption and signature keys for emails.
    
    .. seealso::
       Source - https://tools.ietf.org/html/rfc4880
    """

    class PublicKeyAlgorithms(Enum):
        rsa_encrypt_or_sign_hac = 1
        rsa_encrypt_only_hac = 2
        rsa_sign_only_hac = 3
        elgamal_encrypt_only_elgamal_hac = 16
        dsa_digital_signature_algorithm_fips_hac = 17
        reserved_for_elliptic_curve = 18
        reserved_for_ecdsa = 19
        reserved_formerly_elgamal_encrypt_or_sign_ = 20
        reserved_for_diffie_hellman_x_as_defined_for_ietf_s_mime = 21
        private_experimental_algorithm_00 = 100
        private_experimental_algorithm_01 = 101
        private_experimental_algorithm_02 = 102
        private_experimental_algorithm_03 = 103
        private_experimental_algorithm_04 = 104
        private_experimental_algorithm_05 = 105
        private_experimental_algorithm_06 = 106
        private_experimental_algorithm_07 = 107
        private_experimental_algorithm_08 = 108
        private_experimental_algorithm_09 = 109
        private_experimental_algorithm_10 = 110

    class ServerFlags(Enum):
        no_modify = 128

    class KeyFlags(Enum):
        this_key_may_be_used_to_certify_other_keys = 1
        this_key_may_be_used_to_sign_data = 2
        this_key_may_be_used_to_encrypt_communications = 4
        this_key_may_be_used_to_encrypt_storage = 8
        the_private_component_of_this_key_may_have_been_split_by_a_secret_sharing_mechanism = 16
        this_key_may_be_used_for_authentication = 32
        the_private_component_of_this_key_may_be_in_the_possession_of_more_than_one_person = 128

    class CompressionAlgorithms(Enum):
        uncompressed = 0
        zib = 1
        zlib = 2
        bzip = 3
        private_experimental_algorithm_00 = 100
        private_experimental_algorithm_01 = 101
        private_experimental_algorithm_02 = 102
        private_experimental_algorithm_03 = 103
        private_experimental_algorithm_04 = 104
        private_experimental_algorithm_05 = 105
        private_experimental_algorithm_06 = 106
        private_experimental_algorithm_07 = 107
        private_experimental_algorithm_08 = 108
        private_experimental_algorithm_09 = 109
        private_experimental_algorithm_10 = 110

    class PacketTags(Enum):
        reserved_a_packet_tag_must_not_have_this_value = 0
        public_key_encrypted_session_key_packet = 1
        signature_packet = 2
        symmetric_key_encrypted_session_key_packet = 3
        one_pass_signature_packet = 4
        secret_key_packet = 5
        public_key_packet = 6
        secret_subkey_packet = 7
        compressed_data_packet = 8
        symmetrically_encrypted_data_packet = 9
        marker_packet = 10
        literal_data_packet = 11
        trust_packet = 12
        user_id_packet = 13
        public_subkey_packet = 14
        user_attribute_packet = 17
        sym_encrypted_and_integrity_protected_data_packet = 18
        modification_detection_code_packet = 19
        private_or_experimental_values_0 = 60
        private_or_experimental_values_1 = 61
        private_or_experimental_values_2 = 62
        private_or_experimental_values_3 = 63

    class RevocationCodes(Enum):
        no_reason_specified_key_revocations_or_cert_revocations = 0
        key_is_superseded_key_revocations = 1
        key_material_has_been_compromised_key_revocations = 2
        key_is_retired_and_no_longer_used_key_revocations = 3
        user_id_information_is_no_longer_valid_cert_revocations = 32
        private_use_1 = 100
        private_use_2 = 101
        private_use_3 = 102
        private_use_4 = 103
        private_use_11 = 110

    class HashAlgorithms(Enum):
        md5 = 1
        sha1 = 2
        ripemd160 = 3
        reserved4 = 4
        reserved5 = 5
        reserved6 = 6
        reserved7 = 7
        sha256 = 8
        sha384 = 9
        sha512 = 10
        sha224 = 11
        private_experimental_algorithm_00 = 100
        private_experimental_algorithm_01 = 101
        private_experimental_algorithm_02 = 102
        private_experimental_algorithm_03 = 103
        private_experimental_algorithm_04 = 104
        private_experimental_algorithm_05 = 105
        private_experimental_algorithm_06 = 106
        private_experimental_algorithm_07 = 107
        private_experimental_algorithm_08 = 108
        private_experimental_algorithm_09 = 109
        private_experimental_algorithm_10 = 110

    class SymmetricKeyAlgorithm(Enum):
        plain = 0
        idea = 1
        triple_des = 2
        cast5 = 3
        blowfisch = 4
        reserved5 = 5
        reserved6 = 6
        aes_128 = 7
        aes_192 = 8
        aes_256 = 9
        twofish_256 = 10
        private_experimental_algorithm_00 = 100
        private_experimental_algorithm_01 = 101
        private_experimental_algorithm_02 = 102
        private_experimental_algorithm_03 = 103
        private_experimental_algorithm_04 = 104
        private_experimental_algorithm_05 = 105
        private_experimental_algorithm_06 = 106
        private_experimental_algorithm_07 = 107
        private_experimental_algorithm_08 = 108
        private_experimental_algorithm_09 = 109
        private_experimental_algorithm_10 = 110

    class SubpacketTypes(Enum):
        reserved0 = 0
        reserved1 = 1
        signature_creation_time = 2
        signature_expiration_time = 3
        exportable_certification = 4
        trust_signature = 5
        regular_expression = 6
        revocable = 7
        reserved8 = 8
        key_expiration_time = 9
        placeholder_for_backward_compatibility = 10
        preferred_symmetric_algorithms = 11
        revocation_key = 12
        reserved13 = 13
        reserved14 = 14
        reserved15 = 15
        issuer = 16
        reserved17 = 17
        reserved18 = 18
        reserved19 = 19
        notation_data = 20
        preferred_hash_algorithms = 21
        preferred_compression_algorithms = 22
        key_server_preferences = 23
        preferred_key_server = 24
        primary_user_id = 25
        policy_uri = 26
        key_flags = 27
        signers_user_id = 28
        reason_for_revocation = 29
        features = 30
        signature_target = 31
        embedded_signature = 32
    SEQ_FIELDS = ["packets"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['packets']['start'] = self._io.pos()
        self.packets = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['packets']:
                self._debug['packets']['arr'] = []
            self._debug['packets']['arr'].append({'start': self._io.pos()})
            _t_packets = OpenpgpMessage.Packet(self._io, self, self._root)
            _t_packets._read()
            self.packets.append(_t_packets)
            self._debug['packets']['arr'][len(self.packets) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['packets']['end'] = self._io.pos()

    class PreferredHashAlgorithms(KaitaiStruct):
        SEQ_FIELDS = ["algorithm"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['algorithm']['start'] = self._io.pos()
            self.algorithm = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['algorithm']:
                    self._debug['algorithm']['arr'] = []
                self._debug['algorithm']['arr'].append({'start': self._io.pos()})
                self.algorithm.append(KaitaiStream.resolve_enum(OpenpgpMessage.HashAlgorithms, self._io.read_u1()))
                self._debug['algorithm']['arr'][len(self.algorithm) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['algorithm']['end'] = self._io.pos()


    class PreferredCompressionAlgorithms(KaitaiStruct):
        SEQ_FIELDS = ["algorithm"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['algorithm']['start'] = self._io.pos()
            self.algorithm = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['algorithm']:
                    self._debug['algorithm']['arr'] = []
                self._debug['algorithm']['arr'].append({'start': self._io.pos()})
                self.algorithm.append(KaitaiStream.resolve_enum(OpenpgpMessage.CompressionAlgorithms, self._io.read_u1()))
                self._debug['algorithm']['arr'][len(self.algorithm) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['algorithm']['end'] = self._io.pos()


    class SignersUserId(KaitaiStruct):
        SEQ_FIELDS = ["user_id"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['user_id']['start'] = self._io.pos()
            self.user_id = (self._io.read_bytes_full()).decode(u"UTF-8")
            self._debug['user_id']['end'] = self._io.pos()


    class SecretKeyPacket(KaitaiStruct):
        SEQ_FIELDS = ["public_key", "string_to_key", "symmetric_encryption_algorithm", "secret_key"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['public_key']['start'] = self._io.pos()
            self.public_key = OpenpgpMessage.PublicKeyPacket(self._io, self, self._root)
            self.public_key._read()
            self._debug['public_key']['end'] = self._io.pos()
            self._debug['string_to_key']['start'] = self._io.pos()
            self.string_to_key = self._io.read_u1()
            self._debug['string_to_key']['end'] = self._io.pos()
            if self.string_to_key >= 254:
                self._debug['symmetric_encryption_algorithm']['start'] = self._io.pos()
                self.symmetric_encryption_algorithm = KaitaiStream.resolve_enum(OpenpgpMessage.SymmetricKeyAlgorithm, self._io.read_u1())
                self._debug['symmetric_encryption_algorithm']['end'] = self._io.pos()

            self._debug['secret_key']['start'] = self._io.pos()
            self.secret_key = self._io.read_bytes_full()
            self._debug['secret_key']['end'] = self._io.pos()


    class KeyServerPreferences(KaitaiStruct):
        SEQ_FIELDS = ["flag"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['flag']['start'] = self._io.pos()
            self.flag = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['flag']:
                    self._debug['flag']['arr'] = []
                self._debug['flag']['arr'].append({'start': self._io.pos()})
                self.flag.append(KaitaiStream.resolve_enum(OpenpgpMessage.ServerFlags, self._io.read_u1()))
                self._debug['flag']['arr'][len(self.flag) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['flag']['end'] = self._io.pos()


    class RegularExpression(KaitaiStruct):
        SEQ_FIELDS = ["regex"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['regex']['start'] = self._io.pos()
            self.regex = (self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
            self._debug['regex']['end'] = self._io.pos()


    class Subpackets(KaitaiStruct):
        SEQ_FIELDS = ["subpacketss"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['subpacketss']['start'] = self._io.pos()
            self.subpacketss = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['subpacketss']:
                    self._debug['subpacketss']['arr'] = []
                self._debug['subpacketss']['arr'].append({'start': self._io.pos()})
                _t_subpacketss = OpenpgpMessage.Subpacket(self._io, self, self._root)
                _t_subpacketss._read()
                self.subpacketss.append(_t_subpacketss)
                self._debug['subpacketss']['arr'][len(self.subpacketss) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['subpacketss']['end'] = self._io.pos()


    class RevocationKey(KaitaiStruct):
        SEQ_FIELDS = ["class", "public_key_algorithm", "fingerprint"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['class']['start'] = self._io.pos()
            self.class = self._io.read_u1()
            self._debug['class']['end'] = self._io.pos()
            self._debug['public_key_algorithm']['start'] = self._io.pos()
            self.public_key_algorithm = KaitaiStream.resolve_enum(OpenpgpMessage.PublicKeyAlgorithms, self._io.read_u1())
            self._debug['public_key_algorithm']['end'] = self._io.pos()
            self._debug['fingerprint']['start'] = self._io.pos()
            self.fingerprint = self._io.read_bytes(20)
            self._debug['fingerprint']['end'] = self._io.pos()


    class UserIdPacket(KaitaiStruct):
        SEQ_FIELDS = ["user_id"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['user_id']['start'] = self._io.pos()
            self.user_id = (self._io.read_bytes_full()).decode(u"UTF-8")
            self._debug['user_id']['end'] = self._io.pos()


    class PolicyUri(KaitaiStruct):
        SEQ_FIELDS = ["uri"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['uri']['start'] = self._io.pos()
            self.uri = (self._io.read_bytes_full()).decode(u"UTF-8")
            self._debug['uri']['end'] = self._io.pos()


    class SignatureTarget(KaitaiStruct):
        SEQ_FIELDS = ["public_key_algorithm", "hash_algorithm", "hash"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['public_key_algorithm']['start'] = self._io.pos()
            self.public_key_algorithm = KaitaiStream.resolve_enum(OpenpgpMessage.PublicKeyAlgorithms, self._io.read_u1())
            self._debug['public_key_algorithm']['end'] = self._io.pos()
            self._debug['hash_algorithm']['start'] = self._io.pos()
            self.hash_algorithm = KaitaiStream.resolve_enum(OpenpgpMessage.HashAlgorithms, self._io.read_u1())
            self._debug['hash_algorithm']['end'] = self._io.pos()
            self._debug['hash']['start'] = self._io.pos()
            self.hash = self._io.read_bytes_full()
            self._debug['hash']['end'] = self._io.pos()


    class KeyFlags(KaitaiStruct):
        SEQ_FIELDS = ["flag"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['flag']['start'] = self._io.pos()
            self.flag = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['flag']:
                    self._debug['flag']['arr'] = []
                self._debug['flag']['arr'].append({'start': self._io.pos()})
                self.flag.append(KaitaiStream.resolve_enum(OpenpgpMessage.KeyFlags, self._io.read_u1()))
                self._debug['flag']['arr'][len(self.flag) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['flag']['end'] = self._io.pos()


    class Features(KaitaiStruct):
        SEQ_FIELDS = ["flags"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_bytes_full()
            self._debug['flags']['end'] = self._io.pos()


    class PrimaryUserId(KaitaiStruct):
        SEQ_FIELDS = ["user_id"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['user_id']['start'] = self._io.pos()
            self.user_id = self._io.read_u1()
            self._debug['user_id']['end'] = self._io.pos()


    class Subpacket(KaitaiStruct):
        SEQ_FIELDS = ["len", "subpacket_type", "content"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = OpenpgpMessage.LenSubpacket(self._io, self, self._root)
            self.len._read()
            self._debug['len']['end'] = self._io.pos()
            self._debug['subpacket_type']['start'] = self._io.pos()
            self.subpacket_type = KaitaiStream.resolve_enum(OpenpgpMessage.SubpacketTypes, self._io.read_u1())
            self._debug['subpacket_type']['end'] = self._io.pos()
            self._debug['content']['start'] = self._io.pos()
            _on = self.subpacket_type
            if _on == OpenpgpMessage.SubpacketTypes.preferred_key_server:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.PreferredKeyServer(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.issuer:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.Issuer(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.revocable:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.Revocable(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.signature_target:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.SignatureTarget(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.regular_expression:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.RegularExpression(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.exportable_certification:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.ExportableCertification(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.reason_for_revocation:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.ReasonForRevocation(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.key_server_preferences:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.KeyServerPreferences(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.signature_creation_time:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.SignatureCreationTime(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.preferred_hash_algorithms:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.PreferredHashAlgorithms(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.trust_signature:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.TrustSignature(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.key_expiration_time:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.KeyExpirationTime(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.key_flags:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.KeyFlags(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.signature_expiration_time:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.SignatureExpirationTime(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.features:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.Features(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.signers_user_id:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.SignersUserId(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.notation_data:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.NotationData(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.revocation_key:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.RevocationKey(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.preferred_compression_algorithms:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.PreferredCompressionAlgorithms(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.policy_uri:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.PolicyUri(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.primary_user_id:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.PrimaryUserId(_io__raw_content, self, self._root)
                self.content._read()
            elif _on == OpenpgpMessage.SubpacketTypes.embedded_signature:
                self._raw_content = self._io.read_bytes((self.len.len - 1))
                _io__raw_content = KaitaiStream(BytesIO(self._raw_content))
                self.content = OpenpgpMessage.EmbeddedSignature(_io__raw_content, self, self._root)
                self.content._read()
            else:
                self.content = self._io.read_bytes((self.len.len - 1))
            self._debug['content']['end'] = self._io.pos()


    class OldPacket(KaitaiStruct):
        SEQ_FIELDS = ["len", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            _on = self._parent.len_type
            if _on == 0:
                self.len = self._io.read_u1()
            elif _on == 1:
                self.len = self._io.read_u2be()
            elif _on == 2:
                self.len = self._io.read_u4be()
            self._debug['len']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self._parent.packet_type_old
            if _on == OpenpgpMessage.PacketTags.public_key_packet:
                self._raw_body = self._io.read_bytes(self.len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = OpenpgpMessage.PublicKeyPacket(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == OpenpgpMessage.PacketTags.public_subkey_packet:
                self._raw_body = self._io.read_bytes(self.len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = OpenpgpMessage.PublicKeyPacket(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == OpenpgpMessage.PacketTags.user_id_packet:
                self._raw_body = self._io.read_bytes(self.len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = OpenpgpMessage.UserIdPacket(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == OpenpgpMessage.PacketTags.signature_packet:
                self._raw_body = self._io.read_bytes(self.len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = OpenpgpMessage.SignaturePacket(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == OpenpgpMessage.PacketTags.secret_subkey_packet:
                self._raw_body = self._io.read_bytes(self.len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = OpenpgpMessage.PublicKeyPacket(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == OpenpgpMessage.PacketTags.secret_key_packet:
                self._raw_body = self._io.read_bytes(self.len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = OpenpgpMessage.SecretKeyPacket(_io__raw_body, self, self._root)
                self.body._read()
            else:
                self.body = self._io.read_bytes(self.len)
            self._debug['body']['end'] = self._io.pos()


    class Issuer(KaitaiStruct):
        SEQ_FIELDS = ["keyid"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['keyid']['start'] = self._io.pos()
            self.keyid = self._io.read_u8be()
            self._debug['keyid']['end'] = self._io.pos()


    class ExportableCertification(KaitaiStruct):
        SEQ_FIELDS = ["exportable"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['exportable']['start'] = self._io.pos()
            self.exportable = self._io.read_u1()
            self._debug['exportable']['end'] = self._io.pos()


    class SignatureExpirationTime(KaitaiStruct):
        SEQ_FIELDS = ["time"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['time']['start'] = self._io.pos()
            self.time = self._io.read_u4be()
            self._debug['time']['end'] = self._io.pos()


    class SignatureCreationTime(KaitaiStruct):
        SEQ_FIELDS = ["time"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['time']['start'] = self._io.pos()
            self.time = self._io.read_u4be()
            self._debug['time']['end'] = self._io.pos()


    class SignaturePacket(KaitaiStruct):
        SEQ_FIELDS = ["version", "signature_type", "public_key_algorithm", "hash_algorithm", "len_hashed_subpacket", "hashed_subpackets", "len_unhashed_subpacket", "unhashed_subpackets", "left_signed_hash", "rsa_n", "signature"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u1()
            self._debug['version']['end'] = self._io.pos()
            self._debug['signature_type']['start'] = self._io.pos()
            self.signature_type = self._io.read_u1()
            self._debug['signature_type']['end'] = self._io.pos()
            self._debug['public_key_algorithm']['start'] = self._io.pos()
            self.public_key_algorithm = KaitaiStream.resolve_enum(OpenpgpMessage.PublicKeyAlgorithms, self._io.read_u1())
            self._debug['public_key_algorithm']['end'] = self._io.pos()
            self._debug['hash_algorithm']['start'] = self._io.pos()
            self.hash_algorithm = KaitaiStream.resolve_enum(OpenpgpMessage.HashAlgorithms, self._io.read_u1())
            self._debug['hash_algorithm']['end'] = self._io.pos()
            self._debug['len_hashed_subpacket']['start'] = self._io.pos()
            self.len_hashed_subpacket = self._io.read_u2be()
            self._debug['len_hashed_subpacket']['end'] = self._io.pos()
            self._debug['hashed_subpackets']['start'] = self._io.pos()
            self._raw_hashed_subpackets = self._io.read_bytes(self.len_hashed_subpacket)
            _io__raw_hashed_subpackets = KaitaiStream(BytesIO(self._raw_hashed_subpackets))
            self.hashed_subpackets = OpenpgpMessage.Subpackets(_io__raw_hashed_subpackets, self, self._root)
            self.hashed_subpackets._read()
            self._debug['hashed_subpackets']['end'] = self._io.pos()
            self._debug['len_unhashed_subpacket']['start'] = self._io.pos()
            self.len_unhashed_subpacket = self._io.read_u2be()
            self._debug['len_unhashed_subpacket']['end'] = self._io.pos()
            self._debug['unhashed_subpackets']['start'] = self._io.pos()
            self._raw_unhashed_subpackets = self._io.read_bytes(self.len_unhashed_subpacket)
            _io__raw_unhashed_subpackets = KaitaiStream(BytesIO(self._raw_unhashed_subpackets))
            self.unhashed_subpackets = OpenpgpMessage.Subpackets(_io__raw_unhashed_subpackets, self, self._root)
            self.unhashed_subpackets._read()
            self._debug['unhashed_subpackets']['end'] = self._io.pos()
            self._debug['left_signed_hash']['start'] = self._io.pos()
            self.left_signed_hash = self._io.read_u2be()
            self._debug['left_signed_hash']['end'] = self._io.pos()
            self._debug['rsa_n']['start'] = self._io.pos()
            self.rsa_n = self._io.read_u2be()
            self._debug['rsa_n']['end'] = self._io.pos()
            self._debug['signature']['start'] = self._io.pos()
            self.signature = self._io.read_bytes_full()
            self._debug['signature']['end'] = self._io.pos()


    class Revocable(KaitaiStruct):
        SEQ_FIELDS = ["revocable"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['revocable']['start'] = self._io.pos()
            self.revocable = self._io.read_u1()
            self._debug['revocable']['end'] = self._io.pos()


    class EmbeddedSignature(KaitaiStruct):
        SEQ_FIELDS = ["signature_packet"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['signature_packet']['start'] = self._io.pos()
            self.signature_packet = OpenpgpMessage.SignaturePacket(self._io, self, self._root)
            self.signature_packet._read()
            self._debug['signature_packet']['end'] = self._io.pos()


    class PreferredKeyServer(KaitaiStruct):
        SEQ_FIELDS = ["uri"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['uri']['start'] = self._io.pos()
            self.uri = (self._io.read_bytes_full()).decode(u"UTF-8")
            self._debug['uri']['end'] = self._io.pos()


    class ReasonForRevocation(KaitaiStruct):
        SEQ_FIELDS = ["revocation_code", "reason"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['revocation_code']['start'] = self._io.pos()
            self.revocation_code = KaitaiStream.resolve_enum(OpenpgpMessage.RevocationCodes, self._io.read_u1())
            self._debug['revocation_code']['end'] = self._io.pos()
            self._debug['reason']['start'] = self._io.pos()
            self.reason = (self._io.read_bytes_full()).decode(u"UTF-8")
            self._debug['reason']['end'] = self._io.pos()


    class LenSubpacket(KaitaiStruct):
        SEQ_FIELDS = ["first_octet", "second_octet", "scalar"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['first_octet']['start'] = self._io.pos()
            self.first_octet = self._io.read_u1()
            self._debug['first_octet']['end'] = self._io.pos()
            if  ((self.first_octet >= 192) and (self.first_octet < 255)) :
                self._debug['second_octet']['start'] = self._io.pos()
                self.second_octet = self._io.read_u1()
                self._debug['second_octet']['end'] = self._io.pos()

            if self.first_octet == 255:
                self._debug['scalar']['start'] = self._io.pos()
                self.scalar = self._io.read_u4be()
                self._debug['scalar']['end'] = self._io.pos()


        @property
        def len(self):
            if hasattr(self, '_m_len'):
                return self._m_len if hasattr(self, '_m_len') else None

            self._m_len = (self.first_octet if self.first_octet < 192 else (((((self.first_octet - 192) << 8) + self.second_octet) + 192) if  ((self.first_octet >= 192) and (self.first_octet < 255))  else self.scalar))
            return self._m_len if hasattr(self, '_m_len') else None


    class NotationData(KaitaiStruct):
        SEQ_FIELDS = ["flags", "len_name", "len_value", "name", "value"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_bytes(4)
            self._debug['flags']['end'] = self._io.pos()
            self._debug['len_name']['start'] = self._io.pos()
            self.len_name = self._io.read_u2be()
            self._debug['len_name']['end'] = self._io.pos()
            self._debug['len_value']['start'] = self._io.pos()
            self.len_value = self._io.read_u2be()
            self._debug['len_value']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = self._io.read_bytes(self.len_name)
            self._debug['name']['end'] = self._io.pos()
            self._debug['value']['start'] = self._io.pos()
            self.value = self._io.read_bytes(self.len_value)
            self._debug['value']['end'] = self._io.pos()


    class PublicKeyPacket(KaitaiStruct):
        SEQ_FIELDS = ["version", "timestamp", "public_key_algorithm", "len_alg", "rsa_n", "padding", "rsa_e"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u1()
            self._debug['version']['end'] = self._io.pos()
            self._debug['timestamp']['start'] = self._io.pos()
            self.timestamp = self._io.read_u4be()
            self._debug['timestamp']['end'] = self._io.pos()
            self._debug['public_key_algorithm']['start'] = self._io.pos()
            self.public_key_algorithm = KaitaiStream.resolve_enum(OpenpgpMessage.PublicKeyAlgorithms, self._io.read_u1())
            self._debug['public_key_algorithm']['end'] = self._io.pos()
            self._debug['len_alg']['start'] = self._io.pos()
            self.len_alg = self._io.read_u2be()
            self._debug['len_alg']['end'] = self._io.pos()
            self._debug['rsa_n']['start'] = self._io.pos()
            self.rsa_n = self._io.read_bytes(self.len_alg // 8)
            self._debug['rsa_n']['end'] = self._io.pos()
            self._debug['padding']['start'] = self._io.pos()
            self.padding = self._io.read_u2be()
            self._debug['padding']['end'] = self._io.pos()
            self._debug['rsa_e']['start'] = self._io.pos()
            self.rsa_e = self._io.read_bytes(3)
            self._debug['rsa_e']['end'] = self._io.pos()


    class KeyExpirationTime(KaitaiStruct):
        SEQ_FIELDS = ["time"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['time']['start'] = self._io.pos()
            self.time = self._io.read_u4be()
            self._debug['time']['end'] = self._io.pos()


    class Packet(KaitaiStruct):
        SEQ_FIELDS = ["one", "new_packet_format", "packet_type_new", "packet_type_old", "len_type", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['one']['start'] = self._io.pos()
            self.one = self._io.read_bits_int_be(1) != 0
            self._debug['one']['end'] = self._io.pos()
            self._debug['new_packet_format']['start'] = self._io.pos()
            self.new_packet_format = self._io.read_bits_int_be(1) != 0
            self._debug['new_packet_format']['end'] = self._io.pos()
            if self.new_packet_format:
                self._debug['packet_type_new']['start'] = self._io.pos()
                self.packet_type_new = KaitaiStream.resolve_enum(OpenpgpMessage.PacketTags, self._io.read_bits_int_be(6))
                self._debug['packet_type_new']['end'] = self._io.pos()

            if not (self.new_packet_format):
                self._debug['packet_type_old']['start'] = self._io.pos()
                self.packet_type_old = KaitaiStream.resolve_enum(OpenpgpMessage.PacketTags, self._io.read_bits_int_be(4))
                self._debug['packet_type_old']['end'] = self._io.pos()

            if not (self.new_packet_format):
                self._debug['len_type']['start'] = self._io.pos()
                self.len_type = self._io.read_bits_int_be(2)
                self._debug['len_type']['end'] = self._io.pos()

            self._io.align_to_byte()
            self._debug['body']['start'] = self._io.pos()
            _on = self.new_packet_format
            if _on == False:
                self.body = OpenpgpMessage.OldPacket(self._io, self, self._root)
                self.body._read()
            self._debug['body']['end'] = self._io.pos()


    class TrustSignature(KaitaiStruct):
        SEQ_FIELDS = ["level", "amount"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['level']['start'] = self._io.pos()
            self.level = self._io.read_u1()
            self._debug['level']['end'] = self._io.pos()
            self._debug['amount']['start'] = self._io.pos()
            self.amount = self._io.read_u1()
            self._debug['amount']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Code6502(KaitaiStruct):
    """This spec can be used to disassemble raw stream of 6502 CPU machine
    code into individual operations. Each operation includes an opcode
    and, optionally, an argument. Register arguments are part of the
    `opcode` enum.
    """

    class Opcode(Enum):
        brk_impl = 0
        ora_x_ind = 1
        ora_zpg = 5
        asl_zpg = 6
        php_impl = 8
        ora_imm = 9
        asl_a = 10
        ora_abs = 13
        asl_abs = 14
        bpl_rel = 16
        ora_ind_y = 17
        ora_zpg_x = 21
        asl_zpg_x = 22
        clc_impl = 24
        ora_abs_y = 25
        ora_abs_x = 29
        asl_abs_x = 30
        jsr_abs = 32
        and_x_ind = 33
        bit_zpg = 36
        and_zpg = 37
        rol_zpg = 38
        plp_impl = 40
        and_imm = 41
        rol_a = 42
        bit_abs = 44
        and_abs = 45
        rol_abs = 46
        bmi_rel = 48
        and_ind_y = 49
        and_zpg_x = 53
        rol_zpg_x = 54
        sec_impl = 56
        and_abs_y = 57
        and_abs_x = 61
        rol_abs_x = 62
        rti_impl = 64
        eor_x_ind = 65
        eor_zpg = 69
        lsr_zpg = 70
        pha_impl = 72
        eor_imm = 73
        lsr_a = 74
        jmp_abs = 76
        eor_abs = 77
        lsr_abs = 78
        bvc_rel = 80
        eor_ind_y = 81
        eor_zpg_x = 85
        lsr_zpg_x = 86
        cli_impl = 88
        eor_abs_y = 89
        eor_abs_x = 93
        lsr_abs_x = 94
        rts_impl = 96
        adc_x_ind = 97
        adc_zpg = 101
        ror_zpg = 102
        pla_impl = 104
        adc_imm = 105
        ror_a = 106
        jmp_ind = 108
        adc_abs = 109
        ror_abs = 110
        bvs_rel = 112
        adc_ind_y = 113
        adc_zpg_x = 117
        ror_zpg_x = 118
        sei_impl = 120
        adc_abs_y = 121
        adc_abs_x = 125
        ror_abs_x = 126
        sta_x_ind = 129
        sty_zpg = 132
        sta_zpg = 133
        stx_zpg = 134
        dey_impl = 136
        txa_impl = 138
        sty_abs = 140
        sta_abs = 141
        stx_abs = 142
        bcc_rel = 144
        sta_ind_y = 145
        sty_zpg_x = 148
        sta_zpg_x = 149
        stx_zpg_y = 150
        tya_impl = 152
        sta_abs_y = 153
        txs_impl = 154
        sta_abs_x = 157
        ldy_imm = 160
        lda_x_ind = 161
        ldx_imm = 162
        ldy_zpg = 164
        lda_zpg = 165
        ldx_zpg = 166
        tay_impl = 168
        lda_imm = 169
        tax_impl = 170
        ldy_abs = 172
        lda_abs = 173
        ldx_abs = 174
        bcs_rel = 176
        lda_ind_y = 177
        ldy_zpg_x = 180
        lda_zpg_x = 181
        ldx_zpg_y = 182
        clv_impl = 184
        lda_abs_y = 185
        tsx_impl = 186
        ldy_abs_x = 188
        lda_abs_x = 189
        ldx_abs_y = 190
        cpy_imm = 192
        cmp_x_ind = 193
        cpy_zpg = 196
        cmp_zpg = 197
        dec_zpg = 198
        iny_impl = 200
        cmp_imm = 201
        dex_impl = 202
        cpy_abs = 204
        cmp_abs = 205
        dec_abs = 206
        bne_rel = 208
        cmp_ind_y = 209
        cmp_zpg_x = 213
        dec_zpg_x = 214
        cld_impl = 216
        cmp_abs_y = 217
        cmp_abs_x = 221
        dec_abs_x = 222
        cpx_imm = 224
        sbc_x_ind = 225
        cpx_zpg = 228
        sbc_zpg = 229
        inc_zpg = 230
        inx_impl = 232
        sbc_imm = 233
        nop_impl = 234
        cpx_abs = 236
        sbc_abs = 237
        inc_abs = 238
        beq_rel = 240
        sbc_ind_y = 241
        sbc_zpg_x = 245
        inc_zpg_x = 246
        sed_impl = 248
        sbc_abs_y = 249
        sbc_abs_x = 253
        inc_abs_x = 254
    SEQ_FIELDS = ["operations"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['operations']['start'] = self._io.pos()
        self.operations = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['operations']:
                self._debug['operations']['arr'] = []
            self._debug['operations']['arr'].append({'start': self._io.pos()})
            _t_operations = Code6502.Operation(self._io, self, self._root)
            _t_operations._read()
            self.operations.append(_t_operations)
            self._debug['operations']['arr'][len(self.operations) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['operations']['end'] = self._io.pos()

    class Operation(KaitaiStruct):
        SEQ_FIELDS = ["code", "args"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = KaitaiStream.resolve_enum(Code6502.Opcode, self._io.read_u1())
            self._debug['code']['end'] = self._io.pos()
            self._debug['args']['start'] = self._io.pos()
            _on = self.code
            if _on == Code6502.Opcode.bcc_rel:
                self.args = self._io.read_s1()
            elif _on == Code6502.Opcode.ora_ind_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.lda_ind_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cpx_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sta_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sta_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.bcs_rel:
                self.args = self._io.read_s1()
            elif _on == Code6502.Opcode.ldy_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.lsr_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.and_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.adc_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.sta_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.bne_rel:
                self.args = self._io.read_s1()
            elif _on == Code6502.Opcode.lda_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.adc_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.lsr_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.adc_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.sta_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.cpx_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.jmp_ind:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.adc_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.eor_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.eor_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.sta_x_ind:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sbc_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cpy_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.ldx_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.adc_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.bpl_rel:
                self.args = self._io.read_s1()
            elif _on == Code6502.Opcode.ora_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ror_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.adc_ind_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.eor_ind_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.lda_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.bit_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.rol_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sty_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.jsr_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.eor_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.eor_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.lda_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.lda_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.bmi_rel:
                self.args = self._io.read_s1()
            elif _on == Code6502.Opcode.sty_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.adc_x_ind:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.rol_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.stx_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.asl_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.lsr_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ora_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.adc_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.ldy_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.cmp_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.lda_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.bvs_rel:
                self.args = self._io.read_s1()
            elif _on == Code6502.Opcode.lda_x_ind:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cmp_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.inc_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.asl_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.and_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.ldx_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.and_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cpx_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.dec_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ror_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ldx_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.dec_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sbc_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cmp_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.ror_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.inc_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.and_x_ind:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sbc_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.asl_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.eor_x_ind:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ora_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.ldy_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.sbc_x_ind:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.asl_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sbc_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.rol_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.lsr_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.stx_zpg_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ora_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.eor_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.bit_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.ldx_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.ldy_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.jmp_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.beq_rel:
                self.args = self._io.read_s1()
            elif _on == Code6502.Opcode.dec_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.and_ind_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.and_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cmp_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.eor_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sbc_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.cmp_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.sbc_ind_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cmp_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.stx_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.sty_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.cpy_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.dec_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.ror_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.sta_abs_y:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.inc_abs_x:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.lda_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cmp_ind_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cpy_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ldx_zpg_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.sbc_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ora_x_ind:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.rol_zpg_x:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ora_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.sta_ind_y:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.and_abs:
                self.args = self._io.read_u2le()
            elif _on == Code6502.Opcode.and_imm:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.cmp_x_ind:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.ldy_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.inc_zpg:
                self.args = self._io.read_u1()
            elif _on == Code6502.Opcode.bvc_rel:
                self.args = self._io.read_s1()
            elif _on == Code6502.Opcode.ora_zpg:
                self.args = self._io.read_u1()
            self._debug['args']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class DsStore(KaitaiStruct):
    """Apple macOS '.DS_Store' file format.
    
    .. seealso::
       Source - https://en.wikipedia.org/wiki/.DS_Store
       https://metacpan.org/pod/distribution/Mac-Finder-DSStore/DSStoreFormat.pod
       https://0day.work/parsing-the-ds_store-file-format
    """
    SEQ_FIELDS = ["alignment_header", "buddy_allocator_header"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['alignment_header']['start'] = self._io.pos()
        self.alignment_header = self._io.read_bytes(4)
        self._debug['alignment_header']['end'] = self._io.pos()
        if not self.alignment_header == b"\x00\x00\x00\x01":
            raise kaitaistruct.ValidationNotEqualError(b"\x00\x00\x00\x01", self.alignment_header, self._io, u"/seq/0")
        self._debug['buddy_allocator_header']['start'] = self._io.pos()
        self.buddy_allocator_header = DsStore.BuddyAllocatorHeader(self._io, self, self._root)
        self.buddy_allocator_header._read()
        self._debug['buddy_allocator_header']['end'] = self._io.pos()

    class BuddyAllocatorHeader(KaitaiStruct):
        SEQ_FIELDS = ["magic", "ofs_bookkeeping_info_block", "len_bookkeeping_info_block", "copy_ofs_bookkeeping_info_block", "_unnamed4"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x42\x75\x64\x31":
                raise kaitaistruct.ValidationNotEqualError(b"\x42\x75\x64\x31", self.magic, self._io, u"/types/buddy_allocator_header/seq/0")
            self._debug['ofs_bookkeeping_info_block']['start'] = self._io.pos()
            self.ofs_bookkeeping_info_block = self._io.read_u4be()
            self._debug['ofs_bookkeeping_info_block']['end'] = self._io.pos()
            self._debug['len_bookkeeping_info_block']['start'] = self._io.pos()
            self.len_bookkeeping_info_block = self._io.read_u4be()
            self._debug['len_bookkeeping_info_block']['end'] = self._io.pos()
            self._debug['copy_ofs_bookkeeping_info_block']['start'] = self._io.pos()
            self.copy_ofs_bookkeeping_info_block = self._io.read_u4be()
            self._debug['copy_ofs_bookkeeping_info_block']['end'] = self._io.pos()
            self._debug['_unnamed4']['start'] = self._io.pos()
            self._unnamed4 = self._io.read_bytes(16)
            self._debug['_unnamed4']['end'] = self._io.pos()


    class BuddyAllocatorBody(KaitaiStruct):
        SEQ_FIELDS = ["num_blocks", "_unnamed1", "block_addresses", "num_directories", "directory_entries", "free_lists"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_blocks']['start'] = self._io.pos()
            self.num_blocks = self._io.read_u4be()
            self._debug['num_blocks']['end'] = self._io.pos()
            self._debug['_unnamed1']['start'] = self._io.pos()
            self._unnamed1 = self._io.read_bytes(4)
            self._debug['_unnamed1']['end'] = self._io.pos()
            self._debug['block_addresses']['start'] = self._io.pos()
            self.block_addresses = [None] * (self.num_block_addresses)
            for i in range(self.num_block_addresses):
                if not 'arr' in self._debug['block_addresses']:
                    self._debug['block_addresses']['arr'] = []
                self._debug['block_addresses']['arr'].append({'start': self._io.pos()})
                _t_block_addresses = DsStore.BuddyAllocatorBody.BlockDescriptor(self._io, self, self._root)
                _t_block_addresses._read()
                self.block_addresses[i] = _t_block_addresses
                self._debug['block_addresses']['arr'][i]['end'] = self._io.pos()

            self._debug['block_addresses']['end'] = self._io.pos()
            self._debug['num_directories']['start'] = self._io.pos()
            self.num_directories = self._io.read_u4be()
            self._debug['num_directories']['end'] = self._io.pos()
            self._debug['directory_entries']['start'] = self._io.pos()
            self.directory_entries = [None] * (self.num_directories)
            for i in range(self.num_directories):
                if not 'arr' in self._debug['directory_entries']:
                    self._debug['directory_entries']['arr'] = []
                self._debug['directory_entries']['arr'].append({'start': self._io.pos()})
                _t_directory_entries = DsStore.BuddyAllocatorBody.DirectoryEntry(self._io, self, self._root)
                _t_directory_entries._read()
                self.directory_entries[i] = _t_directory_entries
                self._debug['directory_entries']['arr'][i]['end'] = self._io.pos()

            self._debug['directory_entries']['end'] = self._io.pos()
            self._debug['free_lists']['start'] = self._io.pos()
            self.free_lists = [None] * (self.num_free_lists)
            for i in range(self.num_free_lists):
                if not 'arr' in self._debug['free_lists']:
                    self._debug['free_lists']['arr'] = []
                self._debug['free_lists']['arr'].append({'start': self._io.pos()})
                _t_free_lists = DsStore.BuddyAllocatorBody.FreeList(self._io, self, self._root)
                _t_free_lists._read()
                self.free_lists[i] = _t_free_lists
                self._debug['free_lists']['arr'][i]['end'] = self._io.pos()

            self._debug['free_lists']['end'] = self._io.pos()

        class BlockDescriptor(KaitaiStruct):
            SEQ_FIELDS = ["address_raw"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['address_raw']['start'] = self._io.pos()
                self.address_raw = self._io.read_u4be()
                self._debug['address_raw']['end'] = self._io.pos()

            @property
            def offset(self):
                if hasattr(self, '_m_offset'):
                    return self._m_offset if hasattr(self, '_m_offset') else None

                self._m_offset = ((self.address_raw & ~(self._root.block_address_mask)) + 4)
                return self._m_offset if hasattr(self, '_m_offset') else None

            @property
            def size(self):
                if hasattr(self, '_m_size'):
                    return self._m_size if hasattr(self, '_m_size') else None

                self._m_size = ((1 << self.address_raw) & self._root.block_address_mask)
                return self._m_size if hasattr(self, '_m_size') else None


        class DirectoryEntry(KaitaiStruct):
            SEQ_FIELDS = ["len_name", "name", "block_id"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['len_name']['start'] = self._io.pos()
                self.len_name = self._io.read_u1()
                self._debug['len_name']['end'] = self._io.pos()
                self._debug['name']['start'] = self._io.pos()
                self.name = (self._io.read_bytes(self.len_name)).decode(u"UTF-8")
                self._debug['name']['end'] = self._io.pos()
                self._debug['block_id']['start'] = self._io.pos()
                self.block_id = self._io.read_u4be()
                self._debug['block_id']['end'] = self._io.pos()


        class FreeList(KaitaiStruct):
            SEQ_FIELDS = ["counter", "offsets"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['counter']['start'] = self._io.pos()
                self.counter = self._io.read_u4be()
                self._debug['counter']['end'] = self._io.pos()
                self._debug['offsets']['start'] = self._io.pos()
                self.offsets = [None] * (self.counter)
                for i in range(self.counter):
                    if not 'arr' in self._debug['offsets']:
                        self._debug['offsets']['arr'] = []
                    self._debug['offsets']['arr'].append({'start': self._io.pos()})
                    self.offsets[i] = self._io.read_u4be()
                    self._debug['offsets']['arr'][i]['end'] = self._io.pos()

                self._debug['offsets']['end'] = self._io.pos()


        @property
        def num_block_addresses(self):
            if hasattr(self, '_m_num_block_addresses'):
                return self._m_num_block_addresses if hasattr(self, '_m_num_block_addresses') else None

            self._m_num_block_addresses = 256
            return self._m_num_block_addresses if hasattr(self, '_m_num_block_addresses') else None

        @property
        def num_free_lists(self):
            if hasattr(self, '_m_num_free_lists'):
                return self._m_num_free_lists if hasattr(self, '_m_num_free_lists') else None

            self._m_num_free_lists = 32
            return self._m_num_free_lists if hasattr(self, '_m_num_free_lists') else None

        @property
        def directories(self):
            """Master blocks of the different B-trees."""
            if hasattr(self, '_m_directories'):
                return self._m_directories if hasattr(self, '_m_directories') else None

            io = self._root._io
            self._debug['_m_directories']['start'] = io.pos()
            self._m_directories = [None] * (self.num_directories)
            for i in range(self.num_directories):
                if not 'arr' in self._debug['_m_directories']:
                    self._debug['_m_directories']['arr'] = []
                self._debug['_m_directories']['arr'].append({'start': io.pos()})
                _t__m_directories = DsStore.MasterBlockRef(i, io, self, self._root)
                _t__m_directories._read()
                self._m_directories[i] = _t__m_directories
                self._debug['_m_directories']['arr'][i]['end'] = io.pos()

            self._debug['_m_directories']['end'] = io.pos()
            return self._m_directories if hasattr(self, '_m_directories') else None


    class MasterBlockRef(KaitaiStruct):
        SEQ_FIELDS = []
        def __init__(self, idx, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.idx = idx
            self._debug = collections.defaultdict(dict)

        def _read(self):
            pass

        class MasterBlock(KaitaiStruct):
            SEQ_FIELDS = ["block_id", "num_internal_nodes", "num_records", "num_nodes", "_unnamed4"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['block_id']['start'] = self._io.pos()
                self.block_id = self._io.read_u4be()
                self._debug['block_id']['end'] = self._io.pos()
                self._debug['num_internal_nodes']['start'] = self._io.pos()
                self.num_internal_nodes = self._io.read_u4be()
                self._debug['num_internal_nodes']['end'] = self._io.pos()
                self._debug['num_records']['start'] = self._io.pos()
                self.num_records = self._io.read_u4be()
                self._debug['num_records']['end'] = self._io.pos()
                self._debug['num_nodes']['start'] = self._io.pos()
                self.num_nodes = self._io.read_u4be()
                self._debug['num_nodes']['end'] = self._io.pos()
                self._debug['_unnamed4']['start'] = self._io.pos()
                self._unnamed4 = self._io.read_u4be()
                self._debug['_unnamed4']['end'] = self._io.pos()

            @property
            def root_block(self):
                if hasattr(self, '_m_root_block'):
                    return self._m_root_block if hasattr(self, '_m_root_block') else None

                io = self._root._io
                _pos = io.pos()
                io.seek(self._root.buddy_allocator_body.block_addresses[self.block_id].offset)
                self._debug['_m_root_block']['start'] = io.pos()
                self._m_root_block = DsStore.Block(io, self, self._root)
                self._m_root_block._read()
                self._debug['_m_root_block']['end'] = io.pos()
                io.seek(_pos)
                return self._m_root_block if hasattr(self, '_m_root_block') else None


        @property
        def master_block(self):
            if hasattr(self, '_m_master_block'):
                return self._m_master_block if hasattr(self, '_m_master_block') else None

            _pos = self._io.pos()
            self._io.seek(self._parent.block_addresses[self._parent.directory_entries[self.idx].block_id].offset)
            self._debug['_m_master_block']['start'] = self._io.pos()
            self._raw__m_master_block = self._io.read_bytes(self._parent.block_addresses[self._parent.directory_entries[self.idx].block_id].size)
            _io__raw__m_master_block = KaitaiStream(BytesIO(self._raw__m_master_block))
            self._m_master_block = DsStore.MasterBlockRef.MasterBlock(_io__raw__m_master_block, self, self._root)
            self._m_master_block._read()
            self._debug['_m_master_block']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_master_block if hasattr(self, '_m_master_block') else None


    class Block(KaitaiStruct):
        SEQ_FIELDS = ["mode", "counter", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['mode']['start'] = self._io.pos()
            self.mode = self._io.read_u4be()
            self._debug['mode']['end'] = self._io.pos()
            self._debug['counter']['start'] = self._io.pos()
            self.counter = self._io.read_u4be()
            self._debug['counter']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            self.data = [None] * (self.counter)
            for i in range(self.counter):
                if not 'arr' in self._debug['data']:
                    self._debug['data']['arr'] = []
                self._debug['data']['arr'].append({'start': self._io.pos()})
                _t_data = DsStore.Block.BlockData(self.mode, self._io, self, self._root)
                _t_data._read()
                self.data[i] = _t_data
                self._debug['data']['arr'][i]['end'] = self._io.pos()

            self._debug['data']['end'] = self._io.pos()

        class BlockData(KaitaiStruct):
            SEQ_FIELDS = ["block_id", "record"]
            def __init__(self, mode, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self.mode = mode
                self._debug = collections.defaultdict(dict)

            def _read(self):
                if self.mode > 0:
                    self._debug['block_id']['start'] = self._io.pos()
                    self.block_id = self._io.read_u4be()
                    self._debug['block_id']['end'] = self._io.pos()

                self._debug['record']['start'] = self._io.pos()
                self.record = DsStore.Block.BlockData.Record(self._io, self, self._root)
                self.record._read()
                self._debug['record']['end'] = self._io.pos()

            class Record(KaitaiStruct):
                SEQ_FIELDS = ["filename", "structure_type", "data_type", "value"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['filename']['start'] = self._io.pos()
                    self.filename = DsStore.Block.BlockData.Record.Ustr(self._io, self, self._root)
                    self.filename._read()
                    self._debug['filename']['end'] = self._io.pos()
                    self._debug['structure_type']['start'] = self._io.pos()
                    self.structure_type = DsStore.Block.BlockData.Record.FourCharCode(self._io, self, self._root)
                    self.structure_type._read()
                    self._debug['structure_type']['end'] = self._io.pos()
                    self._debug['data_type']['start'] = self._io.pos()
                    self.data_type = (self._io.read_bytes(4)).decode(u"UTF-8")
                    self._debug['data_type']['end'] = self._io.pos()
                    self._debug['value']['start'] = self._io.pos()
                    _on = self.data_type
                    if _on == u"long":
                        self.value = self._io.read_u4be()
                    elif _on == u"shor":
                        self.value = self._io.read_u4be()
                    elif _on == u"comp":
                        self.value = self._io.read_u8be()
                    elif _on == u"bool":
                        self.value = self._io.read_u1()
                    elif _on == u"ustr":
                        self.value = DsStore.Block.BlockData.Record.Ustr(self._io, self, self._root)
                        self.value._read()
                    elif _on == u"dutc":
                        self.value = self._io.read_u8be()
                    elif _on == u"type":
                        self.value = DsStore.Block.BlockData.Record.FourCharCode(self._io, self, self._root)
                        self.value._read()
                    elif _on == u"blob":
                        self.value = DsStore.Block.BlockData.Record.RecordBlob(self._io, self, self._root)
                        self.value._read()
                    self._debug['value']['end'] = self._io.pos()

                class RecordBlob(KaitaiStruct):
                    SEQ_FIELDS = ["length", "value"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['length']['start'] = self._io.pos()
                        self.length = self._io.read_u4be()
                        self._debug['length']['end'] = self._io.pos()
                        self._debug['value']['start'] = self._io.pos()
                        self.value = self._io.read_bytes(self.length)
                        self._debug['value']['end'] = self._io.pos()


                class Ustr(KaitaiStruct):
                    SEQ_FIELDS = ["length", "value"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['length']['start'] = self._io.pos()
                        self.length = self._io.read_u4be()
                        self._debug['length']['end'] = self._io.pos()
                        self._debug['value']['start'] = self._io.pos()
                        self.value = (self._io.read_bytes((2 * self.length))).decode(u"UTF-16BE")
                        self._debug['value']['end'] = self._io.pos()


                class FourCharCode(KaitaiStruct):
                    SEQ_FIELDS = ["value"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['value']['start'] = self._io.pos()
                        self.value = (self._io.read_bytes(4)).decode(u"UTF-8")
                        self._debug['value']['end'] = self._io.pos()



            @property
            def block(self):
                if hasattr(self, '_m_block'):
                    return self._m_block if hasattr(self, '_m_block') else None

                if self.mode > 0:
                    io = self._root._io
                    _pos = io.pos()
                    io.seek(self._root.buddy_allocator_body.block_addresses[self.block_id].offset)
                    self._debug['_m_block']['start'] = io.pos()
                    self._m_block = DsStore.Block(io, self, self._root)
                    self._m_block._read()
                    self._debug['_m_block']['end'] = io.pos()
                    io.seek(_pos)

                return self._m_block if hasattr(self, '_m_block') else None


        @property
        def rightmost_block(self):
            """Rightmost child block pointer."""
            if hasattr(self, '_m_rightmost_block'):
                return self._m_rightmost_block if hasattr(self, '_m_rightmost_block') else None

            if self.mode > 0:
                io = self._root._io
                _pos = io.pos()
                io.seek(self._root.buddy_allocator_body.block_addresses[self.mode].offset)
                self._debug['_m_rightmost_block']['start'] = io.pos()
                self._m_rightmost_block = DsStore.Block(io, self, self._root)
                self._m_rightmost_block._read()
                self._debug['_m_rightmost_block']['end'] = io.pos()
                io.seek(_pos)

            return self._m_rightmost_block if hasattr(self, '_m_rightmost_block') else None


    @property
    def buddy_allocator_body(self):
        if hasattr(self, '_m_buddy_allocator_body'):
            return self._m_buddy_allocator_body if hasattr(self, '_m_buddy_allocator_body') else None

        _pos = self._io.pos()
        self._io.seek((self.buddy_allocator_header.ofs_bookkeeping_info_block + 4))
        self._debug['_m_buddy_allocator_body']['start'] = self._io.pos()
        self._raw__m_buddy_allocator_body = self._io.read_bytes(self.buddy_allocator_header.len_bookkeeping_info_block)
        _io__raw__m_buddy_allocator_body = KaitaiStream(BytesIO(self._raw__m_buddy_allocator_body))
        self._m_buddy_allocator_body = DsStore.BuddyAllocatorBody(_io__raw__m_buddy_allocator_body, self, self._root)
        self._m_buddy_allocator_body._read()
        self._debug['_m_buddy_allocator_body']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_buddy_allocator_body if hasattr(self, '_m_buddy_allocator_body') else None

    @property
    def block_address_mask(self):
        """Bitmask used to calculate the position and the size of each block
        of the B-tree from the block addresses.
        """
        if hasattr(self, '_m_block_address_mask'):
            return self._m_block_address_mask if hasattr(self, '_m_block_address_mask') else None

        self._m_block_address_mask = 31
        return self._m_block_address_mask if hasattr(self, '_m_block_address_mask') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Dune2Pak(KaitaiStruct):
    """
    .. seealso::
       Source - http://www.shikadi.net/moddingwiki/PAK_Format_(Westwood)
    """
    SEQ_FIELDS = ["dir"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['dir']['start'] = self._io.pos()
        self._raw_dir = self._io.read_bytes(self.dir_size)
        _io__raw_dir = KaitaiStream(BytesIO(self._raw_dir))
        self.dir = Dune2Pak.Files(_io__raw_dir, self, self._root)
        self.dir._read()
        self._debug['dir']['end'] = self._io.pos()

    class Files(KaitaiStruct):
        SEQ_FIELDS = ["files"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['files']['start'] = self._io.pos()
            self.files = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['files']:
                    self._debug['files']['arr'] = []
                self._debug['files']['arr'].append({'start': self._io.pos()})
                _t_files = Dune2Pak.File(i, self._io, self, self._root)
                _t_files._read()
                self.files.append(_t_files)
                self._debug['files']['arr'][len(self.files) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['files']['end'] = self._io.pos()


    class File(KaitaiStruct):
        SEQ_FIELDS = ["ofs", "file_name"]
        def __init__(self, idx, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.idx = idx
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ofs']['start'] = self._io.pos()
            self.ofs = self._io.read_u4le()
            self._debug['ofs']['end'] = self._io.pos()
            if self.ofs != 0:
                self._debug['file_name']['start'] = self._io.pos()
                self.file_name = (self._io.read_bytes_term(0, False, True, True)).decode(u"ASCII")
                self._debug['file_name']['end'] = self._io.pos()


        @property
        def next_ofs0(self):
            if hasattr(self, '_m_next_ofs0'):
                return self._m_next_ofs0 if hasattr(self, '_m_next_ofs0') else None

            if self.ofs != 0:
                self._m_next_ofs0 = self._root.dir.files[(self.idx + 1)].ofs

            return self._m_next_ofs0 if hasattr(self, '_m_next_ofs0') else None

        @property
        def next_ofs(self):
            if hasattr(self, '_m_next_ofs'):
                return self._m_next_ofs if hasattr(self, '_m_next_ofs') else None

            if self.ofs != 0:
                self._m_next_ofs = (self._root._io.size() if self.next_ofs0 == 0 else self.next_ofs0)

            return self._m_next_ofs if hasattr(self, '_m_next_ofs') else None

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            if self.ofs != 0:
                io = self._root._io
                _pos = io.pos()
                io.seek(self.ofs)
                self._debug['_m_body']['start'] = io.pos()
                self._m_body = io.read_bytes((self.next_ofs - self.ofs))
                self._debug['_m_body']['end'] = io.pos()
                io.seek(_pos)

            return self._m_body if hasattr(self, '_m_body') else None


    @property
    def dir_size(self):
        if hasattr(self, '_m_dir_size'):
            return self._m_dir_size if hasattr(self, '_m_dir_size') else None

        _pos = self._io.pos()
        self._io.seek(0)
        self._debug['_m_dir_size']['start'] = self._io.pos()
        self._m_dir_size = self._io.read_u4le()
        self._debug['_m_dir_size']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_dir_size if hasattr(self, '_m_dir_size') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class FtlDat(KaitaiStruct):
    SEQ_FIELDS = ["num_files", "files"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['num_files']['start'] = self._io.pos()
        self.num_files = self._io.read_u4le()
        self._debug['num_files']['end'] = self._io.pos()
        self._debug['files']['start'] = self._io.pos()
        self.files = [None] * (self.num_files)
        for i in range(self.num_files):
            if not 'arr' in self._debug['files']:
                self._debug['files']['arr'] = []
            self._debug['files']['arr'].append({'start': self._io.pos()})
            _t_files = FtlDat.File(self._io, self, self._root)
            _t_files._read()
            self.files[i] = _t_files
            self._debug['files']['arr'][i]['end'] = self._io.pos()

        self._debug['files']['end'] = self._io.pos()

    class File(KaitaiStruct):
        SEQ_FIELDS = ["ofs_meta"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ofs_meta']['start'] = self._io.pos()
            self.ofs_meta = self._io.read_u4le()
            self._debug['ofs_meta']['end'] = self._io.pos()

        @property
        def meta(self):
            if hasattr(self, '_m_meta'):
                return self._m_meta if hasattr(self, '_m_meta') else None

            if self.ofs_meta != 0:
                _pos = self._io.pos()
                self._io.seek(self.ofs_meta)
                self._debug['_m_meta']['start'] = self._io.pos()
                self._m_meta = FtlDat.Meta(self._io, self, self._root)
                self._m_meta._read()
                self._debug['_m_meta']['end'] = self._io.pos()
                self._io.seek(_pos)

            return self._m_meta if hasattr(self, '_m_meta') else None


    class Meta(KaitaiStruct):
        SEQ_FIELDS = ["len_file", "len_filename", "filename", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_file']['start'] = self._io.pos()
            self.len_file = self._io.read_u4le()
            self._debug['len_file']['end'] = self._io.pos()
            self._debug['len_filename']['start'] = self._io.pos()
            self.len_filename = self._io.read_u4le()
            self._debug['len_filename']['end'] = self._io.pos()
            self._debug['filename']['start'] = self._io.pos()
            self.filename = (self._io.read_bytes(self.len_filename)).decode(u"UTF-8")
            self._debug['filename']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            self.body = self._io.read_bytes(self.len_file)
            self._debug['body']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class MonomakhSaprChg(KaitaiStruct):
    """CHG is a container format file used by
    [MONOMAKH-SAPR](https://www.liraland.com/mono/index.php), a software
    package for analysis & design of reinforced concrete multi-storey
    buildings with arbitrary configuration in plan.
    
    CHG is a simple container, which bundles several project files
    together.
    
    Written and tested by Vladimir Shulzhitskiy, 2017
    """
    SEQ_FIELDS = ["title", "ent"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['title']['start'] = self._io.pos()
        self.title = (self._io.read_bytes(10)).decode(u"ascii")
        self._debug['title']['end'] = self._io.pos()
        self._debug['ent']['start'] = self._io.pos()
        self.ent = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['ent']:
                self._debug['ent']['arr'] = []
            self._debug['ent']['arr'].append({'start': self._io.pos()})
            _t_ent = MonomakhSaprChg.Block(self._io, self, self._root)
            _t_ent._read()
            self.ent.append(_t_ent)
            self._debug['ent']['arr'][len(self.ent) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['ent']['end'] = self._io.pos()

    class Block(KaitaiStruct):
        SEQ_FIELDS = ["header", "file_size", "file"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['header']['start'] = self._io.pos()
            self.header = (self._io.read_bytes(13)).decode(u"ascii")
            self._debug['header']['end'] = self._io.pos()
            self._debug['file_size']['start'] = self._io.pos()
            self.file_size = self._io.read_u8le()
            self._debug['file_size']['end'] = self._io.pos()
            self._debug['file']['start'] = self._io.pos()
            self.file = self._io.read_bytes(self.file_size)
            self._debug['file']['end'] = self._io.pos()






from collections.abc import Sequence as AbstractSequence, Set as AbstractSet
from typing import Generic, TypeVar, Iterable, Iterator, List, Optional, Sequence, Set

T = TypeVar("T")


class LazyIterableSequence(Generic[T], Sequence[T], AbstractSequence):
    """A thread-safe list lazily generated from an iterator"""
    def __init__(self, source: Iterable[T]):
        self._source_iter: Optional[Iterator[T]] = iter(source)
        self._items: List[T] = []

    def _get_next_source_item(self) -> T:
        if self._source_iter is None:
            raise StopIteration()
        return next(self._source_iter)

    def _complete(self):
        """Finishes reading all source items from the input iterator"""
        while self._source_iter is not None:
            try:
                _ = self[len(self._items)]
            except IndexError:
                assert self._source_iter is None

    def __getitem__(self, index: int) -> T:
        while self._source_iter is not None and len(self._items) <= index:
            # we have not yet read enough items
            try:
                self._items.append(next(self._source_iter))
            except StopIteration:
                self._source_iter = None
                break
        return self._items[index]

    def __len__(self) -> int:
        self._complete()
        return len(self._items)

    def __iter__(self) -> Iterator[T]:
        i = 0
        while True:
            try:
                yield self[i]
                i += 1
            except IndexError:
                break


def unique(iterator: Iterator[T], elements: Optional[Set[T]] = None) -> Iterator[T]:
    """Yields all of the unique elements of the input sequence"""
    if elements is None:
        elements = set()
    for t in iterator:
        if t not in elements:
            yield t
            elements.add(t)


class LazyIterableSet(Generic[T], AbstractSet, LazyIterableSequence[T]):
    """
    A collection that is both a set and a sequence that
    """
    def __init__(self, source: Iterable[T]):
        self._set: Set[T] = set()
        super().__init__(unique(iter(source), elements=self._set))

    def __contains__(self, x: object) -> bool:
        return x in self._set



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

from polyfile.kaitai.parsers import rtp_packet
class Rtpdump(KaitaiStruct):
    """rtpdump is a format used by rtptools to record and replay
    rtp data from network capture.
    
    .. seealso::
       Source - https://chromium.googlesource.com/external/webrtc/stable/talk/+/master/media/base/rtpdump.h
    """
    SEQ_FIELDS = ["file_header", "packets"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['file_header']['start'] = self._io.pos()
        self.file_header = Rtpdump.HeaderT(self._io, self, self._root)
        self.file_header._read()
        self._debug['file_header']['end'] = self._io.pos()
        self._debug['packets']['start'] = self._io.pos()
        self.packets = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['packets']:
                self._debug['packets']['arr'] = []
            self._debug['packets']['arr'].append({'start': self._io.pos()})
            _t_packets = Rtpdump.PacketT(self._io, self, self._root)
            _t_packets._read()
            self.packets.append(_t_packets)
            self._debug['packets']['arr'][len(self.packets) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['packets']['end'] = self._io.pos()

    class HeaderT(KaitaiStruct):
        SEQ_FIELDS = ["shebang", "space", "ip", "port", "start_sec", "start_usec", "ip2", "port2", "padding"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['shebang']['start'] = self._io.pos()
            self.shebang = self._io.read_bytes(12)
            self._debug['shebang']['end'] = self._io.pos()
            if not self.shebang == b"\x23\x21\x72\x74\x70\x70\x6C\x61\x79\x31\x2E\x30":
                raise kaitaistruct.ValidationNotEqualError(b"\x23\x21\x72\x74\x70\x70\x6C\x61\x79\x31\x2E\x30", self.shebang, self._io, u"/types/header_t/seq/0")
            self._debug['space']['start'] = self._io.pos()
            self.space = self._io.read_bytes(1)
            self._debug['space']['end'] = self._io.pos()
            if not self.space == b"\x20":
                raise kaitaistruct.ValidationNotEqualError(b"\x20", self.space, self._io, u"/types/header_t/seq/1")
            self._debug['ip']['start'] = self._io.pos()
            self.ip = (self._io.read_bytes_term(47, False, True, True)).decode(u"ascii")
            self._debug['ip']['end'] = self._io.pos()
            self._debug['port']['start'] = self._io.pos()
            self.port = (self._io.read_bytes_term(10, False, True, True)).decode(u"ascii")
            self._debug['port']['end'] = self._io.pos()
            self._debug['start_sec']['start'] = self._io.pos()
            self.start_sec = self._io.read_u4be()
            self._debug['start_sec']['end'] = self._io.pos()
            self._debug['start_usec']['start'] = self._io.pos()
            self.start_usec = self._io.read_u4be()
            self._debug['start_usec']['end'] = self._io.pos()
            self._debug['ip2']['start'] = self._io.pos()
            self.ip2 = self._io.read_u4be()
            self._debug['ip2']['end'] = self._io.pos()
            self._debug['port2']['start'] = self._io.pos()
            self.port2 = self._io.read_u2be()
            self._debug['port2']['end'] = self._io.pos()
            self._debug['padding']['start'] = self._io.pos()
            self.padding = self._io.read_u2be()
            self._debug['padding']['end'] = self._io.pos()


    class PacketT(KaitaiStruct):
        SEQ_FIELDS = ["length", "len_body", "packet_usec", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['length']['start'] = self._io.pos()
            self.length = self._io.read_u2be()
            self._debug['length']['end'] = self._io.pos()
            self._debug['len_body']['start'] = self._io.pos()
            self.len_body = self._io.read_u2be()
            self._debug['len_body']['end'] = self._io.pos()
            self._debug['packet_usec']['start'] = self._io.pos()
            self.packet_usec = self._io.read_u4be()
            self._debug['packet_usec']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            self._raw_body = self._io.read_bytes(self.len_body)
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = rtp_packet.RtpPacket(_io__raw_body)
            self.body._read()
            self._debug['body']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class CpioOldLe(KaitaiStruct):
    SEQ_FIELDS = ["files"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['files']['start'] = self._io.pos()
        self.files = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['files']:
                self._debug['files']['arr'] = []
            self._debug['files']['arr'].append({'start': self._io.pos()})
            _t_files = CpioOldLe.File(self._io, self, self._root)
            _t_files._read()
            self.files.append(_t_files)
            self._debug['files']['arr'][len(self.files) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['files']['end'] = self._io.pos()

    class File(KaitaiStruct):
        SEQ_FIELDS = ["header", "path_name", "string_terminator", "path_name_padding", "file_data", "file_data_padding", "end_of_file_padding"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['header']['start'] = self._io.pos()
            self.header = CpioOldLe.FileHeader(self._io, self, self._root)
            self.header._read()
            self._debug['header']['end'] = self._io.pos()
            self._debug['path_name']['start'] = self._io.pos()
            self.path_name = self._io.read_bytes((self.header.path_name_size - 1))
            self._debug['path_name']['end'] = self._io.pos()
            self._debug['string_terminator']['start'] = self._io.pos()
            self.string_terminator = self._io.read_bytes(1)
            self._debug['string_terminator']['end'] = self._io.pos()
            if not self.string_terminator == b"\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x00", self.string_terminator, self._io, u"/types/file/seq/2")
            if (self.header.path_name_size % 2) == 1:
                self._debug['path_name_padding']['start'] = self._io.pos()
                self.path_name_padding = self._io.read_bytes(1)
                self._debug['path_name_padding']['end'] = self._io.pos()
                if not self.path_name_padding == b"\x00":
                    raise kaitaistruct.ValidationNotEqualError(b"\x00", self.path_name_padding, self._io, u"/types/file/seq/3")

            self._debug['file_data']['start'] = self._io.pos()
            self.file_data = self._io.read_bytes(self.header.file_size.value)
            self._debug['file_data']['end'] = self._io.pos()
            if (self.header.file_size.value % 2) == 1:
                self._debug['file_data_padding']['start'] = self._io.pos()
                self.file_data_padding = self._io.read_bytes(1)
                self._debug['file_data_padding']['end'] = self._io.pos()
                if not self.file_data_padding == b"\x00":
                    raise kaitaistruct.ValidationNotEqualError(b"\x00", self.file_data_padding, self._io, u"/types/file/seq/5")

            if  ((self.path_name == b"\x54\x52\x41\x49\x4C\x45\x52\x21\x21\x21") and (self.header.file_size.value == 0)) :
                self._debug['end_of_file_padding']['start'] = self._io.pos()
                self.end_of_file_padding = self._io.read_bytes_full()
                self._debug['end_of_file_padding']['end'] = self._io.pos()



    class FileHeader(KaitaiStruct):
        SEQ_FIELDS = ["magic", "device_number", "inode_number", "mode", "user_id", "group_id", "number_of_links", "r_device_number", "modification_time", "path_name_size", "file_size"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(2)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\xC7\x71":
                raise kaitaistruct.ValidationNotEqualError(b"\xC7\x71", self.magic, self._io, u"/types/file_header/seq/0")
            self._debug['device_number']['start'] = self._io.pos()
            self.device_number = self._io.read_u2le()
            self._debug['device_number']['end'] = self._io.pos()
            self._debug['inode_number']['start'] = self._io.pos()
            self.inode_number = self._io.read_u2le()
            self._debug['inode_number']['end'] = self._io.pos()
            self._debug['mode']['start'] = self._io.pos()
            self.mode = self._io.read_u2le()
            self._debug['mode']['end'] = self._io.pos()
            self._debug['user_id']['start'] = self._io.pos()
            self.user_id = self._io.read_u2le()
            self._debug['user_id']['end'] = self._io.pos()
            self._debug['group_id']['start'] = self._io.pos()
            self.group_id = self._io.read_u2le()
            self._debug['group_id']['end'] = self._io.pos()
            self._debug['number_of_links']['start'] = self._io.pos()
            self.number_of_links = self._io.read_u2le()
            self._debug['number_of_links']['end'] = self._io.pos()
            self._debug['r_device_number']['start'] = self._io.pos()
            self.r_device_number = self._io.read_u2le()
            self._debug['r_device_number']['end'] = self._io.pos()
            self._debug['modification_time']['start'] = self._io.pos()
            self.modification_time = CpioOldLe.FourByteUnsignedInteger(self._io, self, self._root)
            self.modification_time._read()
            self._debug['modification_time']['end'] = self._io.pos()
            self._debug['path_name_size']['start'] = self._io.pos()
            self.path_name_size = self._io.read_u2le()
            self._debug['path_name_size']['end'] = self._io.pos()
            self._debug['file_size']['start'] = self._io.pos()
            self.file_size = CpioOldLe.FourByteUnsignedInteger(self._io, self, self._root)
            self.file_size._read()
            self._debug['file_size']['end'] = self._io.pos()


    class FourByteUnsignedInteger(KaitaiStruct):
        SEQ_FIELDS = ["most_significant_bits", "least_significant_bits"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['most_significant_bits']['start'] = self._io.pos()
            self.most_significant_bits = self._io.read_u2le()
            self._debug['most_significant_bits']['end'] = self._io.pos()
            self._debug['least_significant_bits']['start'] = self._io.pos()
            self.least_significant_bits = self._io.read_u2le()
            self._debug['least_significant_bits']['end'] = self._io.pos()

        @property
        def value(self):
            if hasattr(self, '_m_value'):
                return self._m_value if hasattr(self, '_m_value') else None

            self._m_value = (self.least_significant_bits + (self.most_significant_bits << 16))
            return self._m_value if hasattr(self, '_m_value') else None






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class ApmPartitionTable(KaitaiStruct):
    """
    .. seealso::
       Source - https://en.wikipedia.org/wiki/Apple_Partition_Map
    """
    SEQ_FIELDS = []
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        pass

    class PartitionEntry(KaitaiStruct):
        SEQ_FIELDS = ["magic", "reserved_1", "number_of_partitions", "partition_start", "partition_size", "partition_name", "partition_type", "data_start", "data_size", "partition_status", "boot_code_start", "boot_code_size", "boot_loader_address", "reserved_2", "boot_code_entry", "reserved_3", "boot_code_cksum", "processor_type"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(2)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x50\x4D":
                raise kaitaistruct.ValidationNotEqualError(b"\x50\x4D", self.magic, self._io, u"/types/partition_entry/seq/0")
            self._debug['reserved_1']['start'] = self._io.pos()
            self.reserved_1 = self._io.read_bytes(2)
            self._debug['reserved_1']['end'] = self._io.pos()
            self._debug['number_of_partitions']['start'] = self._io.pos()
            self.number_of_partitions = self._io.read_u4be()
            self._debug['number_of_partitions']['end'] = self._io.pos()
            self._debug['partition_start']['start'] = self._io.pos()
            self.partition_start = self._io.read_u4be()
            self._debug['partition_start']['end'] = self._io.pos()
            self._debug['partition_size']['start'] = self._io.pos()
            self.partition_size = self._io.read_u4be()
            self._debug['partition_size']['end'] = self._io.pos()
            self._debug['partition_name']['start'] = self._io.pos()
            self.partition_name = (KaitaiStream.bytes_terminate(self._io.read_bytes(32), 0, False)).decode(u"ascii")
            self._debug['partition_name']['end'] = self._io.pos()
            self._debug['partition_type']['start'] = self._io.pos()
            self.partition_type = (KaitaiStream.bytes_terminate(self._io.read_bytes(32), 0, False)).decode(u"ascii")
            self._debug['partition_type']['end'] = self._io.pos()
            self._debug['data_start']['start'] = self._io.pos()
            self.data_start = self._io.read_u4be()
            self._debug['data_start']['end'] = self._io.pos()
            self._debug['data_size']['start'] = self._io.pos()
            self.data_size = self._io.read_u4be()
            self._debug['data_size']['end'] = self._io.pos()
            self._debug['partition_status']['start'] = self._io.pos()
            self.partition_status = self._io.read_u4be()
            self._debug['partition_status']['end'] = self._io.pos()
            self._debug['boot_code_start']['start'] = self._io.pos()
            self.boot_code_start = self._io.read_u4be()
            self._debug['boot_code_start']['end'] = self._io.pos()
            self._debug['boot_code_size']['start'] = self._io.pos()
            self.boot_code_size = self._io.read_u4be()
            self._debug['boot_code_size']['end'] = self._io.pos()
            self._debug['boot_loader_address']['start'] = self._io.pos()
            self.boot_loader_address = self._io.read_u4be()
            self._debug['boot_loader_address']['end'] = self._io.pos()
            self._debug['reserved_2']['start'] = self._io.pos()
            self.reserved_2 = self._io.read_bytes(4)
            self._debug['reserved_2']['end'] = self._io.pos()
            self._debug['boot_code_entry']['start'] = self._io.pos()
            self.boot_code_entry = self._io.read_u4be()
            self._debug['boot_code_entry']['end'] = self._io.pos()
            self._debug['reserved_3']['start'] = self._io.pos()
            self.reserved_3 = self._io.read_bytes(4)
            self._debug['reserved_3']['end'] = self._io.pos()
            self._debug['boot_code_cksum']['start'] = self._io.pos()
            self.boot_code_cksum = self._io.read_u4be()
            self._debug['boot_code_cksum']['end'] = self._io.pos()
            self._debug['processor_type']['start'] = self._io.pos()
            self.processor_type = (KaitaiStream.bytes_terminate(self._io.read_bytes(16), 0, False)).decode(u"ascii")
            self._debug['processor_type']['end'] = self._io.pos()

        @property
        def partition(self):
            if hasattr(self, '_m_partition'):
                return self._m_partition if hasattr(self, '_m_partition') else None

            if (self.partition_status & 1) != 0:
                io = self._root._io
                _pos = io.pos()
                io.seek((self.partition_start * self._root.sector_size))
                self._debug['_m_partition']['start'] = io.pos()
                self._m_partition = io.read_bytes((self.partition_size * self._root.sector_size))
                self._debug['_m_partition']['end'] = io.pos()
                io.seek(_pos)

            return self._m_partition if hasattr(self, '_m_partition') else None

        @property
        def data(self):
            if hasattr(self, '_m_data'):
                return self._m_data if hasattr(self, '_m_data') else None

            io = self._root._io
            _pos = io.pos()
            io.seek((self.data_start * self._root.sector_size))
            self._debug['_m_data']['start'] = io.pos()
            self._m_data = io.read_bytes((self.data_size * self._root.sector_size))
            self._debug['_m_data']['end'] = io.pos()
            io.seek(_pos)
            return self._m_data if hasattr(self, '_m_data') else None

        @property
        def boot_code(self):
            if hasattr(self, '_m_boot_code'):
                return self._m_boot_code if hasattr(self, '_m_boot_code') else None

            io = self._root._io
            _pos = io.pos()
            io.seek((self.boot_code_start * self._root.sector_size))
            self._debug['_m_boot_code']['start'] = io.pos()
            self._m_boot_code = io.read_bytes(self.boot_code_size)
            self._debug['_m_boot_code']['end'] = io.pos()
            io.seek(_pos)
            return self._m_boot_code if hasattr(self, '_m_boot_code') else None


    @property
    def sector_size(self):
        """0x200 (512) bytes for disks, 0x1000 (4096) bytes is not supported by APM
        0x800 (2048) bytes for CDROM
        """
        if hasattr(self, '_m_sector_size'):
            return self._m_sector_size if hasattr(self, '_m_sector_size') else None

        self._m_sector_size = 512
        return self._m_sector_size if hasattr(self, '_m_sector_size') else None

    @property
    def partition_lookup(self):
        """Every partition entry contains the number of partition entries.
        We parse the first entry, to know how many to parse, including the first one.
        No logic is given what to do if other entries have a different number.
        """
        if hasattr(self, '_m_partition_lookup'):
            return self._m_partition_lookup if hasattr(self, '_m_partition_lookup') else None

        io = self._root._io
        _pos = io.pos()
        io.seek(self._root.sector_size)
        self._debug['_m_partition_lookup']['start'] = io.pos()
        self._raw__m_partition_lookup = io.read_bytes(self.sector_size)
        _io__raw__m_partition_lookup = KaitaiStream(BytesIO(self._raw__m_partition_lookup))
        self._m_partition_lookup = ApmPartitionTable.PartitionEntry(_io__raw__m_partition_lookup, self, self._root)
        self._m_partition_lookup._read()
        self._debug['_m_partition_lookup']['end'] = io.pos()
        io.seek(_pos)
        return self._m_partition_lookup if hasattr(self, '_m_partition_lookup') else None

    @property
    def partition_entries(self):
        if hasattr(self, '_m_partition_entries'):
            return self._m_partition_entries if hasattr(self, '_m_partition_entries') else None

        io = self._root._io
        _pos = io.pos()
        io.seek(self._root.sector_size)
        self._debug['_m_partition_entries']['start'] = io.pos()
        self._raw__m_partition_entries = [None] * (self._root.partition_lookup.number_of_partitions)
        self._m_partition_entries = [None] * (self._root.partition_lookup.number_of_partitions)
        for i in range(self._root.partition_lookup.number_of_partitions):
            if not 'arr' in self._debug['_m_partition_entries']:
                self._debug['_m_partition_entries']['arr'] = []
            self._debug['_m_partition_entries']['arr'].append({'start': io.pos()})
            self._raw__m_partition_entries[i] = io.read_bytes(self.sector_size)
            _io__raw__m_partition_entries = KaitaiStream(BytesIO(self._raw__m_partition_entries[i]))
            _t__m_partition_entries = ApmPartitionTable.PartitionEntry(_io__raw__m_partition_entries, self, self._root)
            _t__m_partition_entries._read()
            self._m_partition_entries[i] = _t__m_partition_entries
            self._debug['_m_partition_entries']['arr'][i]['end'] = io.pos()

        self._debug['_m_partition_entries']['end'] = io.pos()
        io.seek(_pos)
        return self._m_partition_entries if hasattr(self, '_m_partition_entries') else None


